from yaml.representer import Representer as YamlRepresenter


class Representer(YamlRepresenter):

    def __init__(self, default_style=None, default_flow_style=False,
                 sort_keys=True, register=None):
        super(Representer, self).__init__(default_style=default_style,
                                          default_flow_style=default_flow_style,
                                          sort_keys=sort_keys)
        self.register = register



import py
import pytest

from unityparser.utils import UnityDocument, UnityDocumentError


class TestUnityDocumentSerialization:

    def test_single_doc_unchanged(self, fixtures, tmpdir):
        base_file_path = py.path.local(fixtures['SingleDoc.asset'])
        doc = UnityDocument.load_yaml(str(base_file_path))
        dumped_file_path = tmpdir.join('SingleDoc.asset')
        doc.dump_yaml(file_path=str(dumped_file_path))

        assert base_file_path.read() == dumped_file_path.read()

    def test_multi_doc_unchanged(self, fixtures, tmpdir):
        base_file_path = py.path.local(fixtures['MultiDoc.asset'])
        doc = UnityDocument.load_yaml(str(base_file_path))
        dumped_file_path = tmpdir.join('MultiDoc.asset')
        doc.dump_yaml(file_path=str(dumped_file_path))

        assert base_file_path.read() == dumped_file_path.read()

    def test_unity_extra_anchor_data(self, fixtures, tmpdir):
        base_file_path = py.path.local(fixtures['UnityExtraAnchorData.prefab'])
        doc = UnityDocument.load_yaml(str(base_file_path))
        dumped_file_path = tmpdir.join('UnityExtraAnchorData.prefab')
        doc.dump_yaml(file_path=str(dumped_file_path))

        assert base_file_path.read() == dumped_file_path.read()


class TestUnityDocumentFilters:

    @pytest.mark.parametrize('class_names, attributes, num_entries', [
        (('Transform', 'MonoBehaviour'), ('m_EditorHideFlags',), 1),
        (('SpriteRenderer',), tuple(), 1),
        (('NonExistingClass',), tuple(), 0),
        (('MonoBehaviour',), ('m_NonExitingAttr',), 0),
        (tuple(), tuple(), 5),
        (tuple(), ('m_Enabled',), 2)
    ])
    def test_filter_entries(self, fixtures, class_names, attributes, num_entries):
        multidoc_path = py.path.local(fixtures['MultiDoc.asset'])
        doc = UnityDocument.load_yaml(str(multidoc_path))
        entries = doc.filter(class_names=class_names, attributes=attributes)
        assert len(entries) == num_entries
        if len(class_names):
            assert set([x.__class__.__name__ for x in entries]) <= set(class_names)
        if len(attributes):
            assert all(map(lambda x: all(map(lambda attr: hasattr(x, attr), attributes)), entries))

    @pytest.mark.parametrize('class_name, attributes', [
        ('MonoBehaviour', ('m_EditorHideFlags',)),
        ('SpriteRenderer', tuple()),
        pytest.param('NonExistingClass', tuple(), marks=pytest.mark.xfail(raises=UnityDocumentError)),
        pytest.param('MonoBehaviour', ('m_NonExitingAttr',), marks=pytest.mark.xfail(raises=UnityDocumentError)),
        pytest.param(None, tuple(), marks=pytest.mark.xfail(raises=UnityDocumentError)),
        pytest.param(None, ('m_Enabled',), marks=pytest.mark.xfail(raises=UnityDocumentError))
    ])
    def test_get_entry(self, fixtures, class_name, attributes):
        multidoc_path = py.path.local(fixtures['MultiDoc.asset'])
        doc = UnityDocument.load_yaml(str(multidoc_path))
        entry = doc.get(class_name=class_name, attributes=attributes)
        if class_name is not None:
            assert entry.__class__.__name__ == class_name
        if len(attributes):
            assert all(map(lambda attr: hasattr(entry, attr), attributes))



import re

from yaml.resolver import BaseResolver


class Resolver(BaseResolver):
    """
    UNITY: Resolver will convert EVERYTHING to str but for empty values which are translated to None.
    This is the safest approach as Unity doesn't follow YAML value conventions.
    """
    pass


# Resolver.add_implicit_resolver(
#     'tag:yaml.org,2002:bool',
#     re.compile(r'''^(?:yes|Yes|YES|no|No|NO
#                     |true|True|TRUE|false|False|FALSE
#                     |on|On|ON|off|Off|OFF)$''', re.X),
#     list('yYnNtTfFoO'))

Resolver.add_implicit_resolver(
    'tag:yaml.org,2002:float',
    re.compile(r'''^(?:[-+]?(?:0|[1-9][0-9]*)\.(?:[1-9]|[0-9][0-9]*[1-9])?)$''', re.X),
    list('-+0123456789.'))

Resolver.add_implicit_resolver(
    'tag:yaml.org,2002:int',
    # UNITY: Restrict to simple integer values
    re.compile(r'''^(?:[-+]?(?:0|[1-9][0-9]*))$''', re.X),
    list('-+0123456789'))

# Resolver.add_implicit_resolver(
#     'tag:yaml.org,2002:merge',
#     re.compile(r'^(?:<<)$'),
#     ['<'])

Resolver.add_implicit_resolver(
    'tag:yaml.org,2002:null',
    # UNITY: Allow only empty values as null
    re.compile(r'''^(?: )$''', re.X),
    [''])

# Resolver.add_implicit_resolver(
#     'tag:yaml.org,2002:timestamp',
#     re.compile(r'''^(?:[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]
#                     |[0-9][0-9][0-9][0-9] -[0-9][0-9]? -[0-9][0-9]?
#                      (?:[Tt]|[ \t]+)[0-9][0-9]?
#                      :[0-9][0-9] :[0-9][0-9] (?:\.[0-9]*)?
#                      (?:[ \t]*(?:Z|[-+][0-9][0-9]?(?::[0-9][0-9])?))?)$''', re.X),
#     list('0123456789'))

# Resolver.add_implicit_resolver(
#     'tag:yaml.org,2002:value',
#     re.compile(r'^(?:=)$'),
#     ['='])

# The following resolver is only for documentation purposes. It cannot work
# because plain scalars cannot start with '!', '&', or '*'.
Resolver.add_implicit_resolver(
    'tag:yaml.org,2002:yaml',
    re.compile(r'^(?:!|&|\*)$'),
    list('!&*'))



import io

from .dumper import UnityDumper
from .errors import UnityDocumentError
from .loader import UnityLoader
from .register import UnityScalarRegister

UNIX_LINE_ENDINGS = '\n'


class UnityDocument:

    def __init__(self, data, newline=None, file_path=None, register=None):
        self.newline = newline
        self.data = data
        self.file_path = file_path
        self.register = register or UnityScalarRegister()

    @property
    def entry(self):
        # as many documents contain a single document entry, this might be handy
        return self.data[0]

    @property
    def entries(self):
        return self.data

    def dump_yaml(self, file_path=None):
        """
        :param file_path: If self.file_path is None, it must be passed
        :type file_path:
        :return:
        :rtype:
        """
        file_path = file_path or self.file_path
        assert_or_raise(file_path is not None, UnityDocumentError("file_path parameter must be passed"))
        with open(file_path, 'w', newline=self.newline) as fp:
            dump_all(self.data, stream=fp, register=self.register)

    @classmethod
    def load_yaml(cls, file_path):
        register = UnityScalarRegister()
        with open(file_path, newline='') as fp:
            data = [d for d in load_all(fp, register)]
            # use document line endings if no mixed lien endings found, else default to linux
            line_endings = UNIX_LINE_ENDINGS if isinstance(fp.newlines, tuple) else fp.newlines
        doc = UnityDocument(data, newline=line_endings, file_path=file_path, register=register)
        return doc

    # region Filtering

    def filter(self, class_names=None, attributes=None):
        """
        Filter a group of entries
        :param class_names: iterable of class names to filter
        :type class_names:
        :param attributes: iterable of attribute names that classes must have to be selected
        :type attributes:
        :return: list entries selected
        :rtype:
        """
        entries = self.entries
        if class_names:
            s_class_names = set(class_names)
            entries = filter(lambda x: x.__class__.__name__ in s_class_names, entries)
        if attributes:
            s_attributes = set(attributes)
            entries = filter(lambda x: s_attributes <= x.get_attrs(), entries)
        return list(entries)

    def get(self, class_name=None, attributes=None):
        """
        Filter a single entry. Only, and at least one must exist else it will except
        :param class_name: a class name to get
        :type class_name:
        :param attributes: iterable of attribute names that define the unique entry
        :type attributes:
        :return: a single entry
        :rtype:
        """
        if class_name:
            t_class_name = (class_name,)
        else:
            t_class_name = tuple()
        entries = self.filter(class_names=t_class_name, attributes=attributes)
        assert_or_raise(len(entries) > 0, UnityDocumentError("get method must return on entry. none found"))
        assert_or_raise(len(entries) == 1, UnityDocumentError("get method must return on entry. multiple found"))
        return entries[0]

    # endregion


def assert_or_raise(condition, exception):
    if not condition:
        raise exception


def load_all(stream, register=None):
    """
    Parse all YAML documents in a stream
    and produce corresponding Python objects.
    """
    loader = UnityLoader(stream, register)
    try:
        while loader.check_data():
            yield loader.get_data()
    finally:
        loader.dispose()


def dump_all(documents, stream=None, default_style=None,
             default_flow_style=False,
             canonical=None, indent=None, width=None,
             allow_unicode=None, line_break=None,
             encoding=None, explicit_start=None, explicit_end=None,
             version=None, tags=None, sort_keys=True, register=None):
    """
    Serialize a sequence of Python objects into a YAML stream.
    If stream is None, return the produced string instead.
    """
    getvalue = None
    if stream is None:
        if encoding is None:
            stream = io.StringIO()
        else:
            stream = io.BytesIO()
        getvalue = stream.getvalue
    dumper = UnityDumper(stream, default_style=default_style,
                         default_flow_style=default_flow_style,
                         canonical=canonical, indent=indent, width=width,
                         allow_unicode=allow_unicode, line_break=line_break,
                         encoding=encoding, version=version, tags=tags,
                         explicit_start=explicit_start,
                         explicit_end=explicit_end,
                         sort_keys=sort_keys, register=register)
    try:
        dumper.open()
        for data in documents:
            dumper.represent(data)
        dumper.close()
    finally:
        dumper.dispose()
    if getvalue:
        return getvalue()



from yaml.emitter import (
    Emitter as YamlEmitter,
    EmitterError)
from yaml.events import (
    DocumentStartEvent,
    StreamEndEvent,
    AliasEvent,
    ScalarEvent,
    CollectionStartEvent,
    SequenceStartEvent,
    MappingStartEvent)


class Emitter(YamlEmitter):

    def expect_document_start(self, first=False):
        if isinstance(self.event, DocumentStartEvent):
            # UNITY: omit document end separator
            # if (self.event.version or self.event.tags) and self.open_ended:
            #     self.write_indicator('...', True)
            #     self.write_indent()
            # UNITY: version only written before first document start
            if self.event.version and first:
                version_text = self.prepare_version(self.event.version)
                self.write_version_directive(version_text)
            self.tag_prefixes = self.DEFAULT_TAG_PREFIXES.copy()
            if self.event.tags:
                handles = sorted(self.event.tags.keys())
                for handle in handles:
                    prefix = self.event.tags[handle]
                    self.tag_prefixes[prefix] = handle
                    # UNITY: tag only written before first document start
                    if first:
                        handle_text = self.prepare_tag_handle(handle)
                        prefix_text = self.prepare_tag_prefix(prefix)
                        self.write_tag_directive(handle_text, prefix_text)
            implicit = (first and not self.event.explicit and not self.canonical
                        and not self.event.version and not self.event.tags
                        and not self.check_empty_document())
            if not implicit:
                self.write_indent()
                self.write_indicator('---', True)
                if self.canonical:
                    self.write_indent()
            self.state = self.expect_document_root
        elif isinstance(self.event, StreamEndEvent):
            # UNITY: omit document end separator
            # if self.open_ended:
            #     self.write_indicator('...', True)
            #     self.write_indent()
            self.write_stream_end()
            self.state = self.expect_nothing
        else:
            raise EmitterError("expected DocumentStartEvent, but got %s"
                               % self.event)

    def expect_node(self, root=False, sequence=False, mapping=False,
                    simple_key=False):
        self.root_context = root
        self.sequence_context = sequence
        self.mapping_context = mapping
        self.simple_key_context = simple_key
        if isinstance(self.event, AliasEvent):
            self.expect_alias()
        elif isinstance(self.event, (ScalarEvent, CollectionStartEvent)):
            # UNITY: tag and anchor appearance order swithced
            self.process_tag()
            self.process_anchor('&')
            self.process_anchor_extra()
            if isinstance(self.event, ScalarEvent):
                self.expect_scalar()
            elif isinstance(self.event, SequenceStartEvent):
                if self.flow_level or self.canonical or self.event.flow_style \
                        or self.check_empty_sequence():
                    self.expect_flow_sequence()
                else:
                    self.expect_block_sequence()
            elif isinstance(self.event, MappingStartEvent):
                if self.flow_level or self.canonical or self.event.flow_style \
                        or self.check_empty_mapping():
                    self.expect_flow_mapping()
                else:
                    self.expect_block_mapping()
        else:
            raise EmitterError("expected NodeEvent, but got %s" % self.event)

    def write_plain(self, text, split=True):
        if self.root_context:
            self.open_ended = True
        # UNITY: empty values contain a whitespace before line break
        # if not text:
        #     return
        if not self.whitespace:
            data = ' '
            self.column += len(data)
            if self.encoding:
                data = data.encode(self.encoding)
            self.stream.write(data)
        self.whitespace = False
        self.indention = False
        spaces = False
        breaks = False
        start = end = 0
        while end <= len(text):
            ch = None
            if end < len(text):
                ch = text[end]
            if spaces:
                if ch != ' ':
                    if start+1 == end and self.column > self.best_width and split:
                        self.write_indent()
                        self.whitespace = False
                        self.indention = False
                    else:
                        data = text[start:end]
                        self.column += len(data)
                        if self.encoding:
                            data = data.encode(self.encoding)
                        self.stream.write(data)
                    start = end
            elif breaks:
                if ch not in '\n\x85\u2028\u2029':
                    if text[start] == '\n':
                        self.write_line_break()
                    for br in text[start:end]:
                        if br == '\n':
                            self.write_line_break()
                        else:
                            self.write_line_break(br)
                    self.write_indent()
                    self.whitespace = False
                    self.indention = False
                    start = end
            else:
                if ch is None or ch in ' \n\x85\u2028\u2029':
                    data = text[start:end]
                    self.column += len(data)
                    if self.encoding:
                        data = data.encode(self.encoding)
                    self.stream.write(data)
                    start = end
            if ch is not None:
                spaces = (ch == ' ')
                breaks = (ch in '\n\x85\u2028\u2029')
            end += 1

    def write_double_quoted(self, text, split=True):
        self.write_indicator('"', True)
        start = end = 0
        while end <= len(text):
            ch = None
            if end < len(text):
                ch = text[end]
            if ch is None or ch in '"\\\x85\u2028\u2029\uFEFF' \
                    or not ('\x20' <= ch <= '\x7E'
                        or (self.allow_unicode
                            and ('\xA0' <= ch <= '\uD7FF'
                                or '\uE000' <= ch <= '\uFFFD'))):
                if start < end:
                    data = text[start:end]
                    self.column += len(data)
                    if self.encoding:
                        data = data.encode(self.encoding)
                    self.stream.write(data)
                    start = end
                if ch is not None:
                    if ch in self.ESCAPE_REPLACEMENTS:
                        data = '\\'+self.ESCAPE_REPLACEMENTS[ch]
                    elif ch <= '\xFF':
                        data = '\\x%02X' % ord(ch)
                    elif ch <= '\uFFFF':
                        data = '\\u%04X' % ord(ch)
                    else:
                        data = '\\U%08X' % ord(ch)
                    self.column += len(data)
                    if self.encoding:
                        data = data.encode(self.encoding)
                    self.stream.write(data)
                    start = end+1
            if 0 < end < len(text)-1 and (ch == ' ' or start >= end)    \
                    and self.column+(end-start) > self.best_width and split:
                # UNITY: do not append \ at the end of each split line
                data = text[start:end]
                if start < end:
                    start = end
                self.column += len(data)
                if self.encoding:
                    data = data.encode(self.encoding)
                self.stream.write(data)
                self.write_indent()
                self.whitespace = False
                self.indention = False
                if text[start] == ' ':
                    # UNITY: skip leading whitespace on new line
                    self.column += 1
                    start += 1
            end += 1
        self.write_indicator('"', False)

    def process_anchor_extra(self):
        if self.event.anchor is None or self.event.anchor not in self.extra_anchor_data:
            return
        self.write_indicator(self.extra_anchor_data[self.event.anchor], False)



import py
import pytest

from unityparser.utils import UnityDocument


class TestInvertedScalarLoading:

    def test_inverted_scalar_loading_ok(self, fixtures, tmpdir):
        base_file_path = py.path.local(fixtures['InvertedScalar.dll.meta'])
        doc = UnityDocument.load_yaml(str(base_file_path))
        assert doc.entry['PluginImporter']['platformData'][0]['first']['Any'] is None




class UnityScalarRegister:
    def __init__(self):
        self.map = {}

    def pop(self, value):
        ptr_id = id(value)
        style = self.map.get(ptr_id, None)
        if style is not None:
            style = style[1]
            del self.map[ptr_id]
        return style

    def set(self, value, style):
        ptr_id = id(value)
        assert ptr_id not in self.map, "Duplicated ptr_id ({}) in UnityScalarRegister".format(ptr_id)
        self.map[ptr_id] = (value, style)



from pathlib import Path

import pytest

from unityparser.constants import UnityClassIdMap


@pytest.fixture(scope='session')
def fixtures():
    """
    :return: Available fixtures. Key is the name of the file, value the path to it.
    :rtype: Dict[str,str]
    """
    fixtures_folder = Path(__file__).parent / Path('fixtures')
    fixtures_dict = dict(map(lambda x: (x.name, str(x.resolve())), fixtures_folder.rglob('*')))
    yield fixtures_dict


@pytest.fixture(scope='function', autouse=True)
def reset_unity_class_id_map():
    """
    Provide a clean cache map for every run
    :return:
    :rtype:
    """
    UnityClassIdMap.reset()



import re
from argparse import ArgumentParser
from multiprocessing import Pool, Manager, Process
from pathlib import Path

from .utils import UnityDocument

YAML_HEADER = '%YAML'


class UnityProjectTester:
    """
    Class to run tests on a given Unity project folder
    """
    AVAILABLE_COMMANDS = ('test_no_yaml_is_modified',)

    def __init__(self):
        self.options = None

    def run(self):
        top_parser = ArgumentParser()
        subparser = top_parser.add_subparsers()
        subparser.required = True
        for cmd in UnityProjectTester.AVAILABLE_COMMANDS:
            fn = getattr(self, cmd)
            parser = subparser.add_parser(cmd, help=fn.__doc__)
            parser.set_defaults(func=fn)
        top_parser.add_argument('project_path', help='Path to the Unity project folder')
        top_parser.add_argument('--exclude',
                                help='Exclude regexp when searching project files. Can be specified multiple times.',
                                default=None,
                                action='append')
        top_parser.add_argument('--keep-changes',
                                help='If a file changes after serialization, do not revert the changes.',
                                default=False,
                                action='store_true')
        top_parser.add_argument('--dry-run',
                                help='Dont\'t modify.',
                                default=False,
                                action='store_true')
        top_parser.add_argument('--suffixes',
                                help='Look for files with suffixes in this comma-separated list(default: ".asset,.unity,.prefab").',
                                default='.asset,.unity,.prefab',
                                action='store_true')
        try:
            self.options = top_parser.parse_args()
        except TypeError:
            top_parser.print_help()
            return 2
        # run given function
        self.options.func()

    def test_no_yaml_is_modified(self):
        """
        Recurse the whole project folder looking for serializable files, load and save them all, and check that
        there are no modifications
        """
        if self.options.dry_run:
            print("Dry-run mode enabled: YAMLs won't be dumped.")
            if self.options.keep_changes:
                print("Keep changes mode will not have any effect during dry run.")
        elif self.options.keep_changes:
            print("Keep changes mode enabled: Changes to files will be kept.")

        project_path = Path(self.options.project_path)
        suffixes = [f".{s.strip('.')}" for s in self.options.suffixes.split(',')]
        asset_file_paths = [p for p in project_path.rglob('*') if p.suffix in suffixes]
        print("Found {} '.asset' files".format(len(asset_file_paths)))

        def is_path_included(path):
            # compare regexp against absolute path
            return not any(rexp.search(str(path.resolve())) for rexp in rexps)

        if self.options.exclude is not None:
            rexps = [re.compile(rexp) for rexp in self.options.exclude]
            valid_file_paths = [p for p in filter(is_path_included, asset_file_paths)]
            print("Excluded {} files".format(len(asset_file_paths) - len(valid_file_paths)))
        else:
            valid_file_paths = asset_file_paths

        file_results = []
        with Manager() as manager:
            print_queue = manager.Queue()
            diff_list = manager.list()
            queue_process = Process(target=UnityProjectTester.read_output, args=(print_queue,))
            queue_process.start()
            with Pool() as pool:
                for f in valid_file_paths:
                    async_res = pool.apply_async(UnityProjectTester.open_and_save,
                                                 (f, print_queue, diff_list, self.options.keep_changes,
                                                  self.options.dry_run))
                    file_results.append((f, async_res))
                pool.close()
                pool.join()
            # signal end of queue with None token
            print_queue.put(None)
            queue_process.join()

            error_results = list(filter(lambda r: not r[1].successful(), file_results))
            if len(error_results):
                # raise the first exception
                file_path, result = error_results[0]
                print("Python process evaluating file {} failed with the following exception:".format(
                    file_path.resolve()), flush=True)
                result.get()
            if len(diff_list):
                print("{} files are different now:".format(len(diff_list)))
                print('\n'.join([str(f.resolve()) for f in diff_list]))

    @staticmethod
    def read_output(print_queue):
        msg = print_queue.get()
        while msg is not None:
            print(msg, flush=True)
            msg = print_queue.get()

    @staticmethod
    def open_and_save(asset_file_path, print_queue, diff_list, keep_changes=False, dry_run=False):
        # check YAML version header, save original content
        with open(str(asset_file_path), 'rb') as fp:
            header = fp.read(len(YAML_HEADER))
            try:
                is_yaml_file = header.decode('utf-8') == YAML_HEADER
            except UnicodeDecodeError:
                is_yaml_file = False
            finally:
                if not is_yaml_file:
                    print_queue.put("Ignoring non-yaml file {}".format(asset_file_path))
                    return
                else:
                    fp.seek(0)
                    print_queue.put("Processing {}".format(asset_file_path))
            a_file_content = fp.read()
        doc = UnityDocument.load_yaml(str(asset_file_path))

        if dry_run:
            return

        try:
            doc.dump_yaml()
            with open(str(asset_file_path), 'rb') as fp:
                b_file_content = fp.read()

            # compare
            if a_file_content != b_file_content:
                diff_list.append(asset_file_path)
                if not keep_changes:
                    with open(str(asset_file_path), 'wb') as fp:
                        fp.write(a_file_content)

        except Exception:
            with open(str(asset_file_path), 'wb') as fp:
                fp.write(a_file_content)
            raise


if __name__ == '__main__':
    # None is considered successful
    code = UnityProjectTester().run() or 0
    exit(code)



from yaml.composer import Composer as YamlComposer, ComposerError


class Composer(YamlComposer):

    def compose_document(self):
        # Drop the DOCUMENT-START event.
        self.get_event()

        # UNITY: used to store data after the anchor
        self.extra_anchor_data = {}

        # Compose the root node.
        node = self.compose_node(None, None)

        # Drop the DOCUMENT-END event.
        self.get_event()

        # UNITY: prevent reset anchors after document end so we can access them on constructors
        # self.anchors = {}
        return node

    def get_anchor_from_node(self, node):
        for k, v in self.anchors.items():
            if node == v:
                return k
        raise ComposerError("Expected anchor to be present for node")

    def get_extra_anchor_data_from_node(self, anchor):
        if anchor in self.extra_anchor_data:
            return self.extra_anchor_data[anchor]
        return ''



import collections.abc
from .constants import OrderedFlowDict

from yaml.constructor import Constructor as YamlConstructor, ConstructorError
from yaml.nodes import MappingNode


class uniqstr(str):
    def __new__(cls, content):
        return super().__new__(cls, content)


class Constructor(YamlConstructor):

    def __init__(self, register=None):
        super(Constructor, self).__init__()
        self.register = register

    def construct_mapping(self, node, deep=False):
        if not isinstance(node, MappingNode):
            raise ConstructorError(None, None,
                                   "expected a mapping node, but found %s" % node.id,
                                   node.start_mark)
        # UNITY: dict has to be ordered to reproduce nested maps, also save flow style
        mapping = OrderedFlowDict()
        mapping.set_flow_style(node.flow_style)
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, collections.abc.Hashable):
                raise ConstructorError("while constructing a mapping", node.start_mark,
                                       "found unhashable key", key_node.start_mark)
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping

    def construct_scalar(self, node):
        value = super().construct_scalar(node)
        if isinstance(value, str) and node.style is not None:
            value = uniqstr(value)
            self.register.set(value, node.style)
        return value



import py
import pytest

from unityparser.utils import UnityDocument, UnityDocumentError


class TestScalarValueTypes:

    def test_types(self, fixtures):
        base_file_path = py.path.local(fixtures['MultipleTypesDoc.asset'])
        doc = UnityDocument.load_yaml(str(base_file_path))
        multi_types = doc.entry

        count_map = {'int': 0, 'str': 0, 'float': 0}

        def evaluate_type(attr, parent_map):
            attr_value = parent_map[attr]
            split_attr = attr.split('_')
            if split_attr[0] == 'scalar':
                expected_type = split_attr[1]
                assert issubclass(type(attr_value), eval(expected_type))
                count_map[expected_type] += 1
            elif split_attr[0] == 'map':
                for k in attr_value:
                    evaluate_type(k, attr_value)

        multi_types_attr_map = multi_types.get_serialized_properties_dict()
        for attribute in multi_types_attr_map:
            evaluate_type(attribute, multi_types_attr_map)

        assert count_map['int'] == 4
        assert count_map['str'] == 8
        assert count_map['float'] == 6

    def test_sum_int_type(self, fixtures):
        base_file_path = py.path.local(fixtures['MultipleTypesDoc.asset'])
        doc = UnityDocument.load_yaml(str(base_file_path))
        multi_types = doc.entry

        multi_types.scalar_int_001 += 1
        assert multi_types.scalar_int_001 == 16



from yaml.scanner import Scanner as YamlScanner
from yaml.tokens import AnchorToken


class Scanner(YamlScanner):

    def __init__(self):
        super().__init__()

        # keep track of previous tokens
        self.prev_tokens = []

    def check_prev_token(self, *choices):
        # Check if the previous token is one of the given types.
        if self.prev_tokens:
            if not choices:
                return True
            for choice in choices:
                if isinstance(self.prev_tokens[-1], choice):
                    return True
        return False

    def get_token(self):
        # Return the next token.
        while self.need_more_tokens():
            self.fetch_more_tokens()
        if self.tokens:
            self.tokens_taken += 1
            # UNITY: keep track of previous tokens
            self.prev_tokens.append(self.tokens.pop(0))
            return self.prev_tokens[-1]

    def fetch_anchor(self):

        # ANCHOR could start a simple key.
        self.save_possible_simple_key()

        # No simple keys after ANCHOR.
        self.allow_simple_key = False

        # Scan and add ANCHOR.
        self.tokens.append(self.scan_anchor(AnchorToken))
        
        # UNITY: look for tokens after the anchor
        self.fetch_extra_anchor_data()
        
    def fetch_extra_anchor_data(self):
        # see if we are at end of line, and it so, move
        # onto parsing the rest of the file.  Otherwise
        # we need to store the extra data on the anchor line
        if self.peek() in '\r\n\x85\u2028\u2029':
            return

        # parse the extra data after the anchor
        extra_anchor_data = ''
        while self.peek() not in '\r\n\x85\u2028\u2029':
            ch = self.peek()
            self.forward()
            extra_anchor_data += ch

        # and now store it for later serialization
        self.extra_anchor_data[self.tokens[-1].value] = extra_anchor_data
        return



from yaml.reader import Reader

from .scanner import Scanner
from .composer import Composer
from .constants import UNITY_TAG_URI, OrderedFlowDict, UnityClassIdMap
from .constructor import Constructor
from .parser import Parser
from .resolver import Resolver


class UpgradeVersionError(Exception):
    def __init__(self, *args, should_upgrade=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.should_upgrade = should_upgrade

    def __str__(self):
        msg = super().__str__()
        if self.should_upgrade:
            msg = msg.strip()
            msg += '. ' if msg[-1] != '.' else ' '
            msg += 'This might be an unsupported new Unity version, consider updating this module.'
        return msg


class LoaderVersionError(UpgradeVersionError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, should_upgrade=True, **kwargs)


class UnityLoader(Reader, Scanner, Parser, Composer, Constructor, Resolver):

    def __init__(self, stream, register=None):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        Constructor.__init__(self, register)
        Resolver.__init__(self)


def construct_unity_class(loader, tag_suffix, node):
    try:
        classid = tag_suffix
        classname = node.value[0][0].value
    except (ValueError, KeyError):
        raise LoaderVersionError('Unknown class id {}.'.format(tag_suffix))
    class_attributes_node = node.value[0][1]
    # use ordered dict
    loader.flatten_mapping(class_attributes_node)

    cls = UnityClassIdMap.get_or_create_class_id(classid, classname)
    anchor = loader.get_anchor_from_node(node)
    extra_anchor_data = loader.get_extra_anchor_data_from_node(anchor)
    value = cls(anchor, extra_anchor_data)
    value.update_dict(loader.construct_mapping(class_attributes_node, deep=True))
    return value


def construct_yaml_map(loader, node):
    data = OrderedFlowDict()
    data.set_flow_style(node.flow_style)
    yield data
    value = loader.construct_mapping(node)
    data.update(value)


UnityLoader.add_constructor('tag:yaml.org,2002:map', construct_yaml_map)
UnityLoader.add_multi_constructor(UNITY_TAG_URI, construct_unity_class)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import re

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    readme = f.read()

with io.open('unityparser/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        f.read(),
        re.MULTILINE
    ).group(1)

requirements = {'base': None, 'test': None, 'ci': None}
for k in requirements:
    with open("requirements/{}.txt".format(k)) as f:
        requirements[k] = list(filter(lambda x: bool(x.strip()) and not x.strip().startswith('-r '), f.read().splitlines()))

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='unityparser',
    version=version,
    description='A python library to parse and dump Unity YAML files',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Ricard Valverde',
    author_email='ricard.valverde@socialpoint.es',
    url='https://github.com/socialpoint-labs/unity-yaml-parser',
    license='MIT License',
    python_requires='>=3.6.0',
    packages=['unityparser'],
    keywords=['unity', 'yaml', 'parser', 'serializer'],
    install_requires=requirements.pop('base'),
    extras_require=requirements,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from yaml.events import StreamEndEvent, DocumentStartEvent, MappingEndEvent
from yaml.parser import Parser as YamlParser, ParserError
from yaml.tokens import DocumentEndToken, StreamEndToken, DocumentStartToken, StreamStartToken, KeyToken, ValueToken, BlockEndToken


class Parser(YamlParser):

    def __init__(self):
        super(Parser, self).__init__()
        self.parsing_inverted_scalar = False

    def parse_document_start(self):

        # Parse any extra document end indicators.
        while self.check_token(DocumentEndToken):
            self.get_token()

        # Parse an explicit document.
        if not self.check_token(StreamEndToken):
            token = self.peek_token()
            start_mark = token.start_mark
            # UNITY: only process directives(version and tags) on the first document
            if self.check_prev_token(StreamStartToken):
                version, tags = self.process_directives()
            else:
                version, tags = self.yaml_version, self.tag_handles.copy() if self.tag_handles else None
            if not self.check_token(DocumentStartToken):
                raise ParserError(None, None,
                                  "expected '<document start>', but found %r"
                                  % self.peek_token().id,
                                  self.peek_token().start_mark)
            token = self.get_token()
            end_mark = token.end_mark
            event = DocumentStartEvent(start_mark, end_mark,
                                       explicit=True, version=version, tags=tags)
            self.states.append(self.parse_document_end)
            self.state = self.parse_document_content
        else:
            # Parse the end of the stream.
            token = self.get_token()
            event = StreamEndEvent(token.start_mark, token.end_mark)
            assert not self.states
            assert not self.marks
            self.state = None
        return event

    def parse_block_mapping_key(self):
        if self.check_token(KeyToken):
            token = self.get_token()
            if not self.check_token(KeyToken, ValueToken, BlockEndToken):
                self.states.append(self.parse_block_mapping_value)
                return self.parse_block_node_or_indentless_sequence()
            else:
                self.state = self.parse_block_mapping_value
                return self.process_empty_scalar(token.end_mark)
        # UNITY: https://github.com/socialpoint-labs/unity-yaml-parser/issues/32
        if not self.check_token(BlockEndToken) and not self.parsing_inverted_scalar:
            if self.check_token(ValueToken):
                self.get_token()
                self.parsing_inverted_scalar = True
                self.states.append(self.parse_block_mapping_value)
                return self.parse_block_node_or_indentless_sequence()
            else:
                token = self.peek_token()
                raise ParserError("while parsing a block mapping", self.marks[-1],
                        "expected <block end>, but found %r" % token.id, token.start_mark)
        token = self.get_token()
        event = MappingEndEvent(token.start_mark, token.end_mark)
        self.state = self.states.pop()
        self.marks.pop()
        self.parsing_inverted_scalar = False
        return event



from yaml.serializer import Serializer

from .constants import UNITY_TAG_URI, UnityClass, OrderedFlowDict
from .constructor import uniqstr
from .emitter import Emitter
from .representer import Representer
from .resolver import Resolver
from .serializer import Serializer

UNITY_TAG = {'!u!': UNITY_TAG_URI}
VERSION = (1, 1)


class UnityDumper(Emitter, Serializer, Representer, Resolver):

    def __init__(self, stream,
                 default_style=None, default_flow_style=False,
                 canonical=None, indent=None, width=None,
                 allow_unicode=None, line_break=None,
                 encoding=None, explicit_start=None, explicit_end=None,
                 version=None, tags=None, sort_keys=True, register=None):
        tags = tags or {}
        tags.update(UNITY_TAG)
        version = version or VERSION
        Emitter.__init__(self, stream, canonical=canonical,
                         indent=indent, width=width,
                         allow_unicode=allow_unicode, line_break=line_break)
        Serializer.__init__(self, encoding=encoding,
                            explicit_start=explicit_start, explicit_end=explicit_end,
                            version=version, tags=tags)
        Representer.__init__(self, default_style=default_style,
                             default_flow_style=default_flow_style,
                             sort_keys=sort_keys, register=register)
        Resolver.__init__(self)


def represent_unity_class(dumper, instance):
    data = {instance.__class__.__name__: instance.get_serialized_properties_dict()}
    node = dumper.represent_mapping('{}{}'.format(UNITY_TAG_URI, instance.__class_id), data, False)
    # UNITY: set MappingNode anchor and all set all it's descendants anchors to None
    dumper.anchors[node] = instance.anchor
    dumper.extra_anchor_data[instance.anchor] = instance.extra_anchor_data
    for key, value in node.value:
        dumper.anchor_node(key)
        dumper.anchor_node(value)
    return node


def represent_ordered_flow_dict(dumper, instance):
    return dumper.represent_mapping(Resolver.DEFAULT_MAPPING_TAG, instance.items(),
                                    flow_style=instance.get_flow_style())


def represent_none(dumper, instance):
    return dumper.represent_scalar('tag:yaml.org,2002:null', '')


def represent_str(dumper, instance):
    style = dumper.register.pop(instance)
    return dumper.represent_scalar('tag:yaml.org,2002:str', instance, style=style)


Representer.add_multi_representer(UnityClass, represent_unity_class)
Representer.add_representer(OrderedFlowDict, represent_ordered_flow_dict)
Representer.add_representer(type(None), represent_none)
Representer.add_representer(uniqstr, represent_str)



from collections import OrderedDict
from copy import copy
from multiprocessing import RLock

lock = RLock()


class UnityClassIdMap:
    # Hold 'id-classname':UnityClass values
    # UnityClass cannot be uniquely identified by id(https://docs.unity3d.com/Manual/ClassIDReference.html)
    # because multiple serialized files in a project can belong to multiple Unity versions and some of them,
    # specially very old versions < 5.X, have different names for the same id(ie. id:1001 Prefab vs PrefabInstance)
    __class_id_map = {}

    @classmethod
    def reset(self):
        with lock:
            UnityClassIdMap.__class_id_map.clear()

    @classmethod
    def get_or_create_class_id(cls, classid, classname):
        with lock:
            k = "{}-{}".format(classid, classname)
            try:
                unity_cls = UnityClassIdMap.__class_id_map[k]
            except KeyError:
                unity_cls = type(classname, (UnityClass,), {'__class_id': classid, '__class_name': classname})
                UnityClassIdMap.__class_id_map[k] = unity_cls
            return unity_cls


class UnityClass:
    __class_id = ''
    __class_name = ''

    def __init__(self, anchor, extra_anchor_data):
        self.anchor = anchor
        self.extra_anchor_data = extra_anchor_data

    def get_attrs(self):
        # get attribute set except those belonging to the Python class
        return set(self.__dict__.keys() - ['anchor', 'extra_anchor_data'])

    def update_dict(self, d):
        # replace and append current object attributes to self dict
        old_d = self.__dict__
        self.__dict__ = d
        self.__dict__.update(old_d)

    def get_serialized_properties_dict(self):
        # return a copy of the objects attributes but the ones we don't want
        d = copy(self.__dict__)
        del d['anchor']
        del d['extra_anchor_data']
        return d


class OrderedFlowDict(OrderedDict):
    """
    OrderdDict that tracks yaml flow_style from the MappingNode which created it
    """

    def set_flow_style(self, flow_style):
        self.flow_style = flow_style

    def get_flow_style(self):
        return getattr(self, 'flow_style', None)


UNITY_TAG_URI = u'tag:unity3d.com,2011:'



from yaml.serializer import Serializer as YamlSerializer
from yaml.events import DocumentStartEvent, DocumentEndEvent

# override serialzier class to store data needed
# for extra data on anchor lines
class Serializer(YamlSerializer):

    def __init__(self, encoding=None,
            explicit_start=None, explicit_end=None, version=None, tags=None):
        super().__init__(encoding=encoding,
            explicit_start=explicit_start, explicit_end=explicit_end, version=version,
            tags=tags)
        self.extra_anchor_data = {}

    def serialize(self, node):
        if self.closed is None:
            raise SerializerError("serializer is not opened")
        elif self.closed:
            raise SerializerError("serializer is closed")
        self.emit(DocumentStartEvent(explicit=self.use_explicit_start,
            version=self.use_version, tags=self.use_tags))
        self.anchor_node(node)
        self.serialize_node(node, None, None)
        self.emit(DocumentEndEvent(explicit=self.use_explicit_end))
        self.serialized_nodes = {}
        self.anchors = {}
        self.extra_anchor_data = {}
        self.last_anchor_id = 0

